#include "rotfl.h"
#include "eval.h"

object_t reg_expr;
object_t reg_env;
object_t reg_erator;
object_t reg_erand;

object_t reg_cont;
object_t reg_val;

object_t reg_disc;

object_t reg_pdisc;

static object_t
env_lookup (hashv_t name, object_t env)
{
  while ( 1 )
    {
      switch ( obj_type (env) )
	{
	case O_ENV_EMPTY:
	  error ("Unbound variable");
	  break;
	case O_ENV_BINDING:
	  if ( o_env_binding_name (env) == name )
	    return o_env_binding_val (env);
	  env = o_env_binding_cdr (env);
	  break;
	default:
	  /* Something went wrong */
	  error ("Uncaught case in env_lookup()");
	  break;
	}
    }
}

static void
eval (void)
{
  switch ( obj_type (reg_expr) )
    {
    case O_QUOTE:
      reg_val = o_quote_val (reg_expr);
      break;
    case O_VAR:
      reg_val = env_lookup (o_var_name (reg_expr), reg_env);
      break;
    case O_PROGN:
      reg_cont = new_cont_eval (o_progn_pares (reg_expr),
				reg_env, reg_cont);
      reg_cont = new_cont_eval (o_progn_primus (reg_expr),
				reg_env, reg_cont);
      reg_val = no_object ();
      break;
    case O_APPLY:
      reg_cont = new_cont_app1 (o_apply_rand (reg_expr),
				reg_env, reg_cont);
      reg_cont = new_cont_eval (o_apply_rator (reg_expr),
				reg_env, reg_cont);
      reg_val = no_object ();
      break;
    case O_LAMBDA:
      reg_val = new_userfunc (o_lambda_param (reg_expr),
			      o_lambda_body (reg_expr), reg_env);
      break;
    case O_IFTHENELSE:
      reg_cont = new_cont_test (o_ifthenelse_thenexpr (reg_expr),
				o_ifthenelse_elseexpr (reg_expr),
				reg_env, reg_cont);
      reg_cont = new_cont_eval (o_ifthenelse_test (reg_expr),
				reg_env, reg_cont);
      reg_val = no_object ();
      break;
    case O_CURENV:
      reg_val = reg_env;
      break;
    default:
      /* Something went wrong */
      error ("Uncaught case in eval()");
      break;
    }
  reg_expr = no_object ();
  reg_env = no_object ();
}

static void
apply (void)
{
  switch ( obj_type (reg_erator) )
    {
    case O_PRIMFUNC:
      apply_primfunc ();
      break;
    case O_PRIMFUNC1:
      apply_primfunc1 ();
      break;
    case O_PRIMFUNC2:
      apply_primfunc2 ();
      break;
    case O_USERFUNC:
      reg_env = new_env_binding (o_userfunc_param (reg_erator),
				 reg_erand, o_userfunc_defenv (reg_erator));
      reg_expr = o_userfunc_body (reg_erator);
      eval ();
      break;
    case O_CONT_FINAL:
    case O_CONT_EVAL:
    case O_CONT_APP1:
    case O_CONT_APP:
    case O_CONT_TEST:
    case O_CONT_REFL:
      reg_val = reg_erand;
      reg_cont = reg_erator;
      break;
    default:
      /* Something went wrong */
      error ("Uncaught case in apply()");
      break;
    }
  reg_erator = no_object ();
  reg_erand = no_object ();
}

static void
discontinue (void)
{
  /* Writing this code took me three days of pulling my hair out, so I
   * don't see why I should make it easy for you to understand. :-) */
  reg_pdisc = reg_disc;
  while ( obj_type (reg_pdisc)==O_DISC_REFL )
    {
      int ttl = o_disc_refl_ttl (reg_pdisc);
      --ttl;
      mod_disc_refl (reg_pdisc, ttl,
		     o_disc_refl_cont (reg_pdisc),
		     o_disc_refl_disc (reg_pdisc));
      reg_pdisc = o_disc_refl_disc (reg_pdisc);
      if ( ttl <= 0 )
	{
	  do
	    {
	      reg_val = new_task (reg_val, reg_cont);
	      reg_cont = o_disc_refl_cont (reg_disc);
	      if ( o_disc_refl_ttl (reg_disc) > 0 )
		reg_cont = new_cont_refl (o_disc_refl_ttl (reg_disc),
					  reg_cont);
	      reg_disc = o_disc_refl_disc (reg_disc);
	    }
	  while ( reg_disc != reg_pdisc );
	}
    }
  reg_pdisc = no_object ();
}

void
invoke (void)
{
  switch ( obj_type (reg_cont) )
    {
    case O_CONT_FINAL:
      error ("Program execution terminated");
      break;
    case O_CONT_EVAL:
      /* reg_val is ignored */
      reg_expr = o_cont_eval_expr (reg_cont);
      reg_env = o_cont_eval_env (reg_cont);
      reg_cont = o_cont_eval_cont (reg_cont);
      eval ();
      break;
    case O_CONT_APP1:
      reg_expr = o_cont_app1_rand (reg_cont);
      reg_env = o_cont_app1_env (reg_cont);
      reg_cont = new_cont_app (reg_val, o_cont_app1_cont (reg_cont));
      eval ();
      break;
    case O_CONT_APP:
      reg_erator = o_cont_app_erator (reg_cont);
      reg_erand = reg_val;
      reg_cont = o_cont_app_cont (reg_cont);
      apply ();
      break;
    case O_CONT_TEST:
      if ( obj_type (reg_val) != O_BOOLEAN || o_boolean_val (reg_val) )
	reg_cont = new_cont_eval (o_cont_test_thenexpr (reg_cont),
				  o_cont_test_env (reg_cont),
				  o_cont_test_cont (reg_cont));
      else
	reg_cont = new_cont_eval (o_cont_test_elseexpr (reg_cont),
				  o_cont_test_env (reg_cont),
				  o_cont_test_cont (reg_cont));
      reg_val = no_object ();
      break;
    case O_CONT_REFL:
      if ( obj_type (reg_val) != O_TASK )
	error ("O_CONT_REFL expected O_TASK value");
      reg_disc = new_disc_refl (o_cont_refl_ttl (reg_cont),
				o_cont_refl_cont (reg_cont),
				reg_disc);
      reg_cont = o_task_cont (reg_val);
      reg_val = o_task_val (reg_val);
      return;  /* Do not discontinue() */
    default:
      /* Something went wrong */
      error ("Uncaught case in invoke() (continuation)");
      break;
    }
  discontinue ();
}
