%{
#include "rotfl.h"

#define YYERROR_VERBOSE
%}

%union {
  hashv_t ident;
  object_t val;
  char boolean_v;
  int integer_v;
  int character_v;
}

%token <ident> TOK_IDENT
%token <boolean_v> TOK_BOOLEAN
%token <integer_v> TOK_INTEGER
%token <character_v> TOK_CHARACTER
%token TOK_CURENV
%token TOK_NULLENV
%token TOK_LPAREN
%token TOK_RPAREN
%token TOK_LBRACE
%token TOK_RBRACE
%token TOK_QUESTION
%token TOK_COLON
%token TOK_SEMICOLON
%token TOK_ARROW

%type <val> expr
%type <val> body
%type <val> test
%type <val> simple
%type <val> atomic

%start main

%%

main:
	expr				{ mainprog = $1; }
;

expr:
	body				{ $$ = $1; }
      | TOK_IDENT TOK_ARROW expr	{ $$ = new_lambda ($1, $3); }
;

body:
	test				{ $$ = $1; }
      | body TOK_SEMICOLON		{ $$ = $1; }
      | body TOK_SEMICOLON test		{ $$ = new_progn ($1, $3); }
;

test:
	simple				{ $$ = $1; }
      | simple TOK_QUESTION simple TOK_COLON simple
					{ $$ = new_ifthenelse ($1, $3, $5); }
;

simple:
	atomic				{ $$ = $1; }
      | simple atomic			{ $$ = new_apply ($1, $2); }
;

atomic:
	TOK_LPAREN TOK_RPAREN		{ $$ = new_quote (new_unit ()); }
      | TOK_BOOLEAN			{ $$ = new_quote (new_boolean ($1)); }
      | TOK_INTEGER			{ $$ = new_quote (new_integer ($1)); }
      | TOK_CHARACTER			{ $$ = new_quote (new_character ($1)); }
      | TOK_CURENV			{ $$ = new_curenv (); }
      | TOK_NULLENV			{ $$ = new_quote (new_env_empty ()); }
      | TOK_IDENT			{ $$ = new_var ($1); }
      | TOK_LPAREN expr TOK_RPAREN	{ $$ = $2; }
      | TOK_LBRACE expr TOK_RBRACE	{ $$ = new_quote ($2); }
;

%%

void
yyerror (char *msg)
{
  error (msg);
}
